#include "sdkconfig.h"

#include "esp_wifi.h"
#include "esp_event_loop.h"
#include "esp_log.h"
#include "esp_system.h"
#include "nvs_flash.h"

#include "freertos/FreeRTOS.h"
#include "freertos/task.h"
#include "freertos/event_groups.h"


#include <mbedtls/net.h>
#include <mbedtls/ssl.h>
#include <mbedtls/entropy.h>
#include <mbedtls/ctr_drbg.h>
#include <mbedtls/debug.h>
#include <mbedtls/platform.h>
// https://tls.mbed.org/api/ssl__server_8c_source.html
//
#include <mbedtls/error.h>

#include <esp_log.h>
#include <string.h>
#include <stdio.h>


#define SERVER_PORT "443"
static const char* TAG = "sslServer";

static char errortext[256];

static void my_debug(void *ctx, int level, const char *file, int line, const char *str) {
	((void) level);
	((void) ctx);
	printf("%s:%04d: %s", file, line, str);
}

void sslServer() {
	ESP_LOGD(TAG, ">> sslServer");
	mbedtls_net_context      server_fd;
	mbedtls_net_context      listen_fd;
	mbedtls_entropy_context  entropy;
	mbedtls_ctr_drbg_context ctr_drbg;
	mbedtls_ssl_context      ssl;
	mbedtls_ssl_config       conf;
	mbedtls_x509_crt         srvcert;
  mbedtls_pk_context       pkey;

	int ret;
	int len;
	char *pers = "ssl_server";
	unsigned char buf[1024];


	mbedtls_net_init(&server_fd);
	mbedtls_net_init(&listen_fd);
	mbedtls_ssl_init(&ssl);
	mbedtls_ssl_config_init(&conf);
	mbedtls_x509_crt_init(&srvcert);
  mbedtls_pk_init(&pkey);
	mbedtls_entropy_init(&entropy);
	mbedtls_ctr_drbg_init(&ctr_drbg);


	mbedtls_ssl_conf_dbg(&conf, my_debug, stdout);

	//mbedtls_debug_set_threshold(4); // Log at error only

    extern const unsigned char cacert_pem_start[] asm("_binary_cacert_pem_start");
    extern const unsigned char cacert_pem_end[]   asm("_binary_cacert_pem_end");

	ret = mbedtls_x509_crt_parse(&srvcert, (const unsigned char *)cacert_pem_start, cacert_pem_end - cacert_pem_start);
	if( ret != 0 ) {
		ESP_LOGE(TAG, " failed\n  !  mbedtls_x509_crt_parse returned %d\n\n", ret );
		return;
	}

    extern const unsigned char prvtkey_pem_start[] asm("_binary_prvtkey_pem_start");
    extern const unsigned char prvtkey_pem_end[]   asm("_binary_prvtkey_pem_end");

	ret =  mbedtls_pk_parse_key(&pkey, (const unsigned char *) prvtkey_pem_start, prvtkey_pem_end - prvtkey_pem_start, NULL, 0);
	if( ret != 0 ) {
		ESP_LOGE(TAG, " failed\n  !  mbedtls_pk_parse_key returned %d\n\n", ret );
		return;
	}

	ret = mbedtls_net_bind(&server_fd, NULL, "9080", MBEDTLS_NET_PROTO_TCP);
	if( ret != 0 ) {
		ESP_LOGE(TAG, " failed\n  ! mbedtls_net_bind returned %d\n\n", ret );
		return;
	}

	ret = mbedtls_ctr_drbg_seed(&ctr_drbg, mbedtls_entropy_func, &entropy, (const unsigned char *) pers, strlen(pers));
	if (ret != 0) {
		ESP_LOGE(TAG, " failed\n  ! mbedtls_ctr_drbg_seed returned %d\n", ret);
		return;
	}

	ret = mbedtls_ssl_config_defaults(&conf, MBEDTLS_SSL_IS_SERVER, MBEDTLS_SSL_TRANSPORT_STREAM, MBEDTLS_SSL_PRESET_DEFAULT);
	if (ret != 0) {
		ESP_LOGE(TAG, " failed\n  ! mbedtls_ssl_config_defaults returned %d\n\n", ret);
		return;
	}

	mbedtls_ssl_conf_rng(&conf, mbedtls_ctr_drbg_random, &ctr_drbg);


	mbedtls_ssl_conf_ca_chain( &conf, srvcert.next, NULL );
	ret = mbedtls_ssl_conf_own_cert( &conf, &srvcert, &pkey );
	if(ret != 0) {
		ESP_LOGE(TAG, " failed\n  ! mbedtls_ssl_conf_own_cert returned %d\n\n", ret );
		return;
	}

	ret = mbedtls_ssl_setup(&ssl, &conf);
	if (ret != 0) {
		mbedtls_strerror(ret, errortext, sizeof(errortext));
		ESP_LOGE(TAG, "error from mbedtls_ssl_setup: %d -%x - %s\n", ret, ret, errortext);
		return;
	}

	while(1) {
		mbedtls_net_free(&listen_fd);
		mbedtls_ssl_session_reset(&ssl);


		//mbedtls_ssl_conf_authmode(&conf, MBEDTLS_SSL_VERIFY_NONE);

		ESP_LOGD(TAG, ">> waiting for accept");
		ret = mbedtls_net_accept( &server_fd, &listen_fd, NULL, 0, NULL );
		if(ret != 0)	{
			ESP_LOGE(TAG, " failed\n  ! mbedtls_net_accept returned %d\n\n", ret );
			return;
		}

		mbedtls_ssl_set_bio(&ssl, &listen_fd, mbedtls_net_send, mbedtls_net_recv, NULL);

		while((ret = mbedtls_ssl_handshake(&ssl)) != 0) {
			if(ret != MBEDTLS_ERR_SSL_WANT_READ && ret != MBEDTLS_ERR_SSL_WANT_WRITE) {
				ESP_LOGE(TAG, " failed\n  ! mbedtls_ssl_handshake returned %d\n\n", ret );
				return;
			}
		}


		len = sizeof(buf);
		ret = mbedtls_ssl_read(&ssl, buf, len);
		if (ret < 0) {
			ESP_LOGE(TAG, "error from read: %d\n", len);
			return;
		}
		if (ret == 0) {
		  ESP_LOGI(TAG, "connection closed");
		}

		printf("Result: [size: %d]\n%.*s\n", ret, ret, buf);
	}

	mbedtls_net_free(&listen_fd);
	mbedtls_net_free(&server_fd);
	mbedtls_ssl_free(&ssl);
	mbedtls_ssl_config_free(&conf);
	mbedtls_ctr_drbg_free(&ctr_drbg);
	mbedtls_entropy_free(&entropy);
	ESP_LOGV(TAG, "All done");
}

/* FreeRTOS event group to signal when we are connected & ready to make a request */
static EventGroupHandle_t wifi_event_group;

/* The event group allows multiple bits for each event,
   but we only care about one event - are we connected
   to the AP with an IP? */
const int CONNECTED_BIT = BIT0;

static esp_err_t event_handler(void *ctx, system_event_t *event)
{
    switch(event->event_id) {
    case SYSTEM_EVENT_STA_START:
        esp_wifi_connect();
        break;
    case SYSTEM_EVENT_STA_GOT_IP:
        xEventGroupSetBits(wifi_event_group, CONNECTED_BIT);
        break;
    case SYSTEM_EVENT_STA_DISCONNECTED:
        /* This is a workaround as ESP32 WiFi libs don't currently
           auto-reassociate. */
        esp_wifi_connect();
        xEventGroupClearBits(wifi_event_group, CONNECTED_BIT);
        break;
    default:
        break;
    }
    return ESP_OK;
}

static void initialise_wifi(void)
{
    tcpip_adapter_init();
    wifi_event_group = xEventGroupCreate();
    ESP_ERROR_CHECK( esp_event_loop_init(event_handler, NULL) );
    wifi_init_config_t cfg = WIFI_INIT_CONFIG_DEFAULT();
    ESP_ERROR_CHECK( esp_wifi_init(&cfg) );
    ESP_ERROR_CHECK( esp_wifi_set_storage(WIFI_STORAGE_RAM) );
    wifi_config_t wifi_config = {
        .sta = {
            .ssid = "CHANGE THIS",
            .password = "CHANGE THIS",
        },
    };
    ESP_LOGI(TAG, "Setting WiFi configuration SSID %s...", wifi_config.sta.ssid);
    ESP_ERROR_CHECK( esp_wifi_set_mode(WIFI_MODE_STA) );
    ESP_ERROR_CHECK( esp_wifi_set_config(ESP_IF_WIFI_STA, &wifi_config) );
    ESP_ERROR_CHECK( esp_wifi_start() );
}

void ssl_server_wrapper(void *ignore)
{
  while (1) {
	xEventGroupWaitBits(wifi_event_group, CONNECTED_BIT,
						false, true, portMAX_DELAY);
	ESP_LOGI(TAG, "Connected to AP");
	sslServer();
	vTaskDelete(NULL);
	ESP_LOGI(TAG, "Done");
  }
}

void app_main()
{
    ESP_ERROR_CHECK( nvs_flash_init() );
    initialise_wifi();

	xTaskCreate(&ssl_server_wrapper, "ssl_task", 8192, NULL, 5, NULL);
}

